#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 8
#endif
#ifndef PFXDIST
   #define PFXDIST 256
#endif
#ifndef PFADIST
   #define PFADIST 256
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGEMV(ATL_CINT M0, ATL_CINT N, const TYPE *A, ATL_CINT lda0,
               const TYPE *X, TYPE *Y)
/*
 * 8x8 unrolled mvt_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N8=(N/8)*8, M=M0+M0, M8=((M0/8)*8)<<1,
            lda=lda0+lda0, lda8=lda*8;
   ATL_INT j;

   for (j=N8; j; j -= 8, A += lda8, Y += 16)
   {
      const float *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda,
                  *A5=A4+lda, *A6=A5+lda, *A7=A6+lda;
      register float ry0=ATL_rzero, iy0=ATL_rzero, ry1=ATL_rzero, iy1=ATL_rzero,
                     ry2=ATL_rzero, iy2=ATL_rzero, ry3=ATL_rzero, iy3=ATL_rzero,
                     ry4=ATL_rzero, iy4=ATL_rzero, ry5=ATL_rzero, iy5=ATL_rzero,
                     ry6=ATL_rzero, iy6=ATL_rzero, ry7=ATL_rzero,
                     iy7=ATL_rzero;
      ATL_INT i;
      prefY(Y+16+16-1);
      for (i=0; i < M8; i += 16)
      {
         const register float ra0_0=A0[i+0], ia0_0=A0[i+1], ra1_0=A0[i+2],
                              ia1_0=A0[i+3], ra2_0=A0[i+4], ia2_0=A0[i+5],
                              ra3_0=A0[i+6], ia3_0=A0[i+7], ra4_0=A0[i+8],
                              ia4_0=A0[i+9], ra5_0=A0[i+10], ia5_0=A0[i+11],
                              ra6_0=A0[i+12], ia6_0=A0[i+13], ra7_0=A0[i+14],
                              ia7_0=A0[i+15], ra0_1=A1[i+0], ia0_1=A1[i+1],
                              ra1_1=A1[i+2], ia1_1=A1[i+3], ra2_1=A1[i+4],
                              ia2_1=A1[i+5], ra3_1=A1[i+6], ia3_1=A1[i+7],
                              ra4_1=A1[i+8], ia4_1=A1[i+9], ra5_1=A1[i+10],
                              ia5_1=A1[i+11], ra6_1=A1[i+12], ia6_1=A1[i+13],
                              ra7_1=A1[i+14], ia7_1=A1[i+15], ra0_2=A2[i+0],
                              ia0_2=A2[i+1], ra1_2=A2[i+2], ia1_2=A2[i+3],
                              ra2_2=A2[i+4], ia2_2=A2[i+5], ra3_2=A2[i+6],
                              ia3_2=A2[i+7], ra4_2=A2[i+8], ia4_2=A2[i+9],
                              ra5_2=A2[i+10], ia5_2=A2[i+11], ra6_2=A2[i+12],
                              ia6_2=A2[i+13], ra7_2=A2[i+14], ia7_2=A2[i+15],
                              ra0_3=A3[i+0], ia0_3=A3[i+1], ra1_3=A3[i+2],
                              ia1_3=A3[i+3], ra2_3=A3[i+4], ia2_3=A3[i+5],
                              ra3_3=A3[i+6], ia3_3=A3[i+7], ra4_3=A3[i+8],
                              ia4_3=A3[i+9], ra5_3=A3[i+10], ia5_3=A3[i+11],
                              ra6_3=A3[i+12], ia6_3=A3[i+13], ra7_3=A3[i+14],
                              ia7_3=A3[i+15], ra0_4=A4[i+0], ia0_4=A4[i+1],
                              ra1_4=A4[i+2], ia1_4=A4[i+3], ra2_4=A4[i+4],
                              ia2_4=A4[i+5], ra3_4=A4[i+6], ia3_4=A4[i+7],
                              ra4_4=A4[i+8], ia4_4=A4[i+9], ra5_4=A4[i+10],
                              ia5_4=A4[i+11], ra6_4=A4[i+12], ia6_4=A4[i+13],
                              ra7_4=A4[i+14], ia7_4=A4[i+15], ra0_5=A5[i+0],
                              ia0_5=A5[i+1], ra1_5=A5[i+2], ia1_5=A5[i+3],
                              ra2_5=A5[i+4], ia2_5=A5[i+5], ra3_5=A5[i+6],
                              ia3_5=A5[i+7], ra4_5=A5[i+8], ia4_5=A5[i+9],
                              ra5_5=A5[i+10], ia5_5=A5[i+11], ra6_5=A5[i+12],
                              ia6_5=A5[i+13], ra7_5=A5[i+14], ia7_5=A5[i+15],
                              ra0_6=A6[i+0], ia0_6=A6[i+1], ra1_6=A6[i+2],
                              ia1_6=A6[i+3], ra2_6=A6[i+4], ia2_6=A6[i+5],
                              ra3_6=A6[i+6], ia3_6=A6[i+7], ra4_6=A6[i+8],
                              ia4_6=A6[i+9], ra5_6=A6[i+10], ia5_6=A6[i+11],
                              ra6_6=A6[i+12], ia6_6=A6[i+13], ra7_6=A6[i+14],
                              ia7_6=A6[i+15], ra0_7=A7[i+0], ia0_7=A7[i+1],
                              ra1_7=A7[i+2], ia1_7=A7[i+3], ra2_7=A7[i+4],
                              ia2_7=A7[i+5], ra3_7=A7[i+6], ia3_7=A7[i+7],
                              ra4_7=A7[i+8], ia4_7=A7[i+9], ra5_7=A7[i+10],
                              ia5_7=A7[i+11], ra6_7=A7[i+12], ia6_7=A7[i+13],
                              ra7_7=A7[i+14], ia7_7=A7[i+15];
         const register float rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                              rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                              rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                              rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                              ix7=X[i+15];

         prefX(X);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry1 += ra0_1 * rx0;
         iy1 += ra0_1 * ix0;
         ry2 += ra0_2 * rx0;
         iy2 += ra0_2 * ix0;
         ry3 += ra0_3 * rx0;
         iy3 += ra0_3 * ix0;
         ry4 += ra0_4 * rx0;
         iy4 += ra0_4 * ix0;
         ry5 += ra0_5 * rx0;
         iy5 += ra0_5 * ix0;
         ry6 += ra0_6 * rx0;
         iy6 += ra0_6 * ix0;
         ry7 += ra0_7 * rx0;
         iy7 += ra0_7 * ix0;
         ry0 += ra1_0 * rx1;
         iy0 += ra1_0 * ix1;
         ry1 += ra1_1 * rx1;
         iy1 += ra1_1 * ix1;
         ry2 += ra1_2 * rx1;
         iy2 += ra1_2 * ix1;
         ry3 += ra1_3 * rx1;
         iy3 += ra1_3 * ix1;
         ry4 += ra1_4 * rx1;
         iy4 += ra1_4 * ix1;
         ry5 += ra1_5 * rx1;
         iy5 += ra1_5 * ix1;
         ry6 += ra1_6 * rx1;
         iy6 += ra1_6 * ix1;
         ry7 += ra1_7 * rx1;
         iy7 += ra1_7 * ix1;
         ry0 += ra2_0 * rx2;
         iy0 += ra2_0 * ix2;
         ry1 += ra2_1 * rx2;
         iy1 += ra2_1 * ix2;
         ry2 += ra2_2 * rx2;
         iy2 += ra2_2 * ix2;
         ry3 += ra2_3 * rx2;
         iy3 += ra2_3 * ix2;
         ry4 += ra2_4 * rx2;
         iy4 += ra2_4 * ix2;
         ry5 += ra2_5 * rx2;
         iy5 += ra2_5 * ix2;
         ry6 += ra2_6 * rx2;
         iy6 += ra2_6 * ix2;
         ry7 += ra2_7 * rx2;
         iy7 += ra2_7 * ix2;
         ry0 += ra3_0 * rx3;
         iy0 += ra3_0 * ix3;
         ry1 += ra3_1 * rx3;
         iy1 += ra3_1 * ix3;
         ry2 += ra3_2 * rx3;
         iy2 += ra3_2 * ix3;
         ry3 += ra3_3 * rx3;
         iy3 += ra3_3 * ix3;
         ry4 += ra3_4 * rx3;
         iy4 += ra3_4 * ix3;
         ry5 += ra3_5 * rx3;
         iy5 += ra3_5 * ix3;
         ry6 += ra3_6 * rx3;
         iy6 += ra3_6 * ix3;
         ry7 += ra3_7 * rx3;
         iy7 += ra3_7 * ix3;
         ry0 += ra4_0 * rx4;
         iy0 += ra4_0 * ix4;
         ry1 += ra4_1 * rx4;
         iy1 += ra4_1 * ix4;
         ry2 += ra4_2 * rx4;
         iy2 += ra4_2 * ix4;
         ry3 += ra4_3 * rx4;
         iy3 += ra4_3 * ix4;
         ry4 += ra4_4 * rx4;
         iy4 += ra4_4 * ix4;
         ry5 += ra4_5 * rx4;
         iy5 += ra4_5 * ix4;
         ry6 += ra4_6 * rx4;
         iy6 += ra4_6 * ix4;
         ry7 += ra4_7 * rx4;
         iy7 += ra4_7 * ix4;
         ry0 += ra5_0 * rx5;
         iy0 += ra5_0 * ix5;
         ry1 += ra5_1 * rx5;
         iy1 += ra5_1 * ix5;
         ry2 += ra5_2 * rx5;
         iy2 += ra5_2 * ix5;
         ry3 += ra5_3 * rx5;
         iy3 += ra5_3 * ix5;
         ry4 += ra5_4 * rx5;
         iy4 += ra5_4 * ix5;
         ry5 += ra5_5 * rx5;
         iy5 += ra5_5 * ix5;
         ry6 += ra5_6 * rx5;
         iy6 += ra5_6 * ix5;
         ry7 += ra5_7 * rx5;
         iy7 += ra5_7 * ix5;
         ry0 += ra6_0 * rx6;
         iy0 += ra6_0 * ix6;
         ry1 += ra6_1 * rx6;
         iy1 += ra6_1 * ix6;
         ry2 += ra6_2 * rx6;
         iy2 += ra6_2 * ix6;
         ry3 += ra6_3 * rx6;
         iy3 += ra6_3 * ix6;
         ry4 += ra6_4 * rx6;
         iy4 += ra6_4 * ix6;
         ry5 += ra6_5 * rx6;
         iy5 += ra6_5 * ix6;
         ry6 += ra6_6 * rx6;
         iy6 += ra6_6 * ix6;
         ry7 += ra6_7 * rx6;
         iy7 += ra6_7 * ix6;
         ry0 += ra7_0 * rx7;
         iy0 += ra7_0 * ix7;
         ry1 += ra7_1 * rx7;
         iy1 += ra7_1 * ix7;
         ry2 += ra7_2 * rx7;
         iy2 += ra7_2 * ix7;
         ry3 += ra7_3 * rx7;
         iy3 += ra7_3 * ix7;
         ry4 += ra7_4 * rx7;
         iy4 += ra7_4 * ix7;
         ry5 += ra7_5 * rx7;
         iy5 += ra7_5 * ix7;
         ry6 += ra7_6 * rx7;
         iy6 += ra7_6 * ix7;
         ry7 += ra7_7 * rx7;
         iy7 += ra7_7 * ix7;

         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry1 -= ia0_1 * ix0;
         iy1 += ia0_1 * rx0;
         prefA(A1);
         ry2 -= ia0_2 * ix0;
         iy2 += ia0_2 * rx0;
         prefA(A2);
         ry3 -= ia0_3 * ix0;
         iy3 += ia0_3 * rx0;
         prefA(A3);
         ry4 -= ia0_4 * ix0;
         iy4 += ia0_4 * rx0;
         prefA(A4);
         ry5 -= ia0_5 * ix0;
         iy5 += ia0_5 * rx0;
         prefA(A5);
         ry6 -= ia0_6 * ix0;
         iy6 += ia0_6 * rx0;
         prefA(A6);
         ry7 -= ia0_7 * ix0;
         iy7 += ia0_7 * rx0;
         prefA(A7);
         ry0 -= ia1_0 * ix1;
         iy0 += ia1_0 * rx1;
         ry1 -= ia1_1 * ix1;
         iy1 += ia1_1 * rx1;
         ry2 -= ia1_2 * ix1;
         iy2 += ia1_2 * rx1;
         ry3 -= ia1_3 * ix1;
         iy3 += ia1_3 * rx1;
         ry4 -= ia1_4 * ix1;
         iy4 += ia1_4 * rx1;
         ry5 -= ia1_5 * ix1;
         iy5 += ia1_5 * rx1;
         ry6 -= ia1_6 * ix1;
         iy6 += ia1_6 * rx1;
         ry7 -= ia1_7 * ix1;
         iy7 += ia1_7 * rx1;
         ry0 -= ia2_0 * ix2;
         iy0 += ia2_0 * rx2;
         ry1 -= ia2_1 * ix2;
         iy1 += ia2_1 * rx2;
         ry2 -= ia2_2 * ix2;
         iy2 += ia2_2 * rx2;
         ry3 -= ia2_3 * ix2;
         iy3 += ia2_3 * rx2;
         ry4 -= ia2_4 * ix2;
         iy4 += ia2_4 * rx2;
         ry5 -= ia2_5 * ix2;
         iy5 += ia2_5 * rx2;
         ry6 -= ia2_6 * ix2;
         iy6 += ia2_6 * rx2;
         ry7 -= ia2_7 * ix2;
         iy7 += ia2_7 * rx2;
         ry0 -= ia3_0 * ix3;
         iy0 += ia3_0 * rx3;
         ry1 -= ia3_1 * ix3;
         iy1 += ia3_1 * rx3;
         ry2 -= ia3_2 * ix3;
         iy2 += ia3_2 * rx3;
         ry3 -= ia3_3 * ix3;
         iy3 += ia3_3 * rx3;
         ry4 -= ia3_4 * ix3;
         iy4 += ia3_4 * rx3;
         ry5 -= ia3_5 * ix3;
         iy5 += ia3_5 * rx3;
         ry6 -= ia3_6 * ix3;
         iy6 += ia3_6 * rx3;
         ry7 -= ia3_7 * ix3;
         iy7 += ia3_7 * rx3;
         ry0 -= ia4_0 * ix4;
         iy0 += ia4_0 * rx4;
         ry1 -= ia4_1 * ix4;
         iy1 += ia4_1 * rx4;
         ry2 -= ia4_2 * ix4;
         iy2 += ia4_2 * rx4;
         ry3 -= ia4_3 * ix4;
         iy3 += ia4_3 * rx4;
         ry4 -= ia4_4 * ix4;
         iy4 += ia4_4 * rx4;
         ry5 -= ia4_5 * ix4;
         iy5 += ia4_5 * rx4;
         ry6 -= ia4_6 * ix4;
         iy6 += ia4_6 * rx4;
         ry7 -= ia4_7 * ix4;
         iy7 += ia4_7 * rx4;
         ry0 -= ia5_0 * ix5;
         iy0 += ia5_0 * rx5;
         ry1 -= ia5_1 * ix5;
         iy1 += ia5_1 * rx5;
         ry2 -= ia5_2 * ix5;
         iy2 += ia5_2 * rx5;
         ry3 -= ia5_3 * ix5;
         iy3 += ia5_3 * rx5;
         ry4 -= ia5_4 * ix5;
         iy4 += ia5_4 * rx5;
         ry5 -= ia5_5 * ix5;
         iy5 += ia5_5 * rx5;
         ry6 -= ia5_6 * ix5;
         iy6 += ia5_6 * rx5;
         ry7 -= ia5_7 * ix5;
         iy7 += ia5_7 * rx5;
         ry0 -= ia6_0 * ix6;
         iy0 += ia6_0 * rx6;
         ry1 -= ia6_1 * ix6;
         iy1 += ia6_1 * rx6;
         ry2 -= ia6_2 * ix6;
         iy2 += ia6_2 * rx6;
         ry3 -= ia6_3 * ix6;
         iy3 += ia6_3 * rx6;
         ry4 -= ia6_4 * ix6;
         iy4 += ia6_4 * rx6;
         ry5 -= ia6_5 * ix6;
         iy5 += ia6_5 * rx6;
         ry6 -= ia6_6 * ix6;
         iy6 += ia6_6 * rx6;
         ry7 -= ia6_7 * ix6;
         iy7 += ia6_7 * rx6;
         ry0 -= ia7_0 * ix7;
         iy0 += ia7_0 * rx7;
         ry1 -= ia7_1 * ix7;
         iy1 += ia7_1 * rx7;
         ry2 -= ia7_2 * ix7;
         iy2 += ia7_2 * rx7;
         ry3 -= ia7_3 * ix7;
         iy3 += ia7_3 * rx7;
         ry4 -= ia7_4 * ix7;
         iy4 += ia7_4 * rx7;
         ry5 -= ia7_5 * ix7;
         iy5 += ia7_5 * rx7;
         ry6 -= ia7_6 * ix7;
         iy6 += ia7_6 * rx7;
         ry7 -= ia7_7 * ix7;
         iy7 += ia7_7 * rx7;
      }
      for (i=M8; i < M; i += 2)
      {
         const register float rx0 = X[i], ix0 = X[i+1];
         const register float ra0_0=A0[i], ia0_0=A0[i+1], ra0_1=A1[i],
                              ia0_1=A1[i+1], ra0_2=A2[i], ia0_2=A2[i+1],
                              ra0_3=A3[i], ia0_3=A3[i+1], ra0_4=A4[i],
                              ia0_4=A4[i+1], ra0_5=A5[i], ia0_5=A5[i+1],
                              ra0_6=A6[i], ia0_6=A6[i+1], ra0_7=A7[i],
                              ia0_7=A7[i+1];
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry1 += ra0_1 * rx0;
         iy1 += ra0_1 * ix0;
         ry2 += ra0_2 * rx0;
         iy2 += ra0_2 * ix0;
         ry3 += ra0_3 * rx0;
         iy3 += ra0_3 * ix0;
         ry4 += ra0_4 * rx0;
         iy4 += ra0_4 * ix0;
         ry5 += ra0_5 * rx0;
         iy5 += ra0_5 * ix0;
         ry6 += ra0_6 * rx0;
         iy6 += ra0_6 * ix0;
         ry7 += ra0_7 * rx0;
         iy7 += ra0_7 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         ry1 -= ia0_1 * ix0;
         iy1 += ia0_1 * rx0;
         ry2 -= ia0_2 * ix0;
         iy2 += ia0_2 * rx0;
         ry3 -= ia0_3 * ix0;
         iy3 += ia0_3 * rx0;
         ry4 -= ia0_4 * ix0;
         iy4 += ia0_4 * rx0;
         ry5 -= ia0_5 * ix0;
         iy5 += ia0_5 * rx0;
         ry6 -= ia0_6 * ix0;
         iy6 += ia0_6 * rx0;
         ry7 -= ia0_7 * ix0;
         iy7 += ia0_7 * rx0;
      }
      #ifdef BETA0
         Y[0] = ry0;
         Y[1] = iy0;
         Y[2] = ry1;
         Y[3] = iy1;
         Y[4] = ry2;
         Y[5] = iy2;
         Y[6] = ry3;
         Y[7] = iy3;
         Y[8] = ry4;
         Y[9] = iy4;
         Y[10] = ry5;
         Y[11] = iy5;
         Y[12] = ry6;
         Y[13] = iy6;
         Y[14] = ry7;
         Y[15] = iy7;
      #else
         Y[0] += ry0;
         Y[1] += iy0;
         Y[2] += ry1;
         Y[3] += iy1;
         Y[4] += ry2;
         Y[5] += iy2;
         Y[6] += ry3;
         Y[7] += iy3;
         Y[8] += ry4;
         Y[9] += iy4;
         Y[10] += ry5;
         Y[11] += iy5;
         Y[12] += ry6;
         Y[13] += iy6;
         Y[14] += ry7;
         Y[15] += iy7;
      #endif
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N8; j; j--, A += lda, Y += 2)
   {
      const float *A0=A;
      register float ry0=ATL_rzero, iy0=ATL_rzero;
      ATL_INT i;
      prefY(Y+16+16-1);
      for (i=0; i < M8; i += 16)
      {
         const register float ra0_0=A0[i+0], ia0_0=A0[i+1], ra1_0=A0[i+2],
                              ia1_0=A0[i+3], ra2_0=A0[i+4], ia2_0=A0[i+5],
                              ra3_0=A0[i+6], ia3_0=A0[i+7], ra4_0=A0[i+8],
                              ia4_0=A0[i+9], ra5_0=A0[i+10], ia5_0=A0[i+11],
                              ra6_0=A0[i+12], ia6_0=A0[i+13], ra7_0=A0[i+14],
                              ia7_0=A0[i+15];
         const register float rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                              rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                              rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                              rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                              ix7=X[i+15];

         prefX(X);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 += ra1_0 * rx1;
         iy0 += ra1_0 * ix1;
         ry0 += ra2_0 * rx2;
         iy0 += ra2_0 * ix2;
         ry0 += ra3_0 * rx3;
         iy0 += ra3_0 * ix3;
         ry0 += ra4_0 * rx4;
         iy0 += ra4_0 * ix4;
         ry0 += ra5_0 * rx5;
         iy0 += ra5_0 * ix5;
         ry0 += ra6_0 * rx6;
         iy0 += ra6_0 * ix6;
         ry0 += ra7_0 * rx7;
         iy0 += ra7_0 * ix7;

         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry0 -= ia1_0 * ix1;
         iy0 += ia1_0 * rx1;
         ry0 -= ia2_0 * ix2;
         iy0 += ia2_0 * rx2;
         ry0 -= ia3_0 * ix3;
         iy0 += ia3_0 * rx3;
         ry0 -= ia4_0 * ix4;
         iy0 += ia4_0 * rx4;
         ry0 -= ia5_0 * ix5;
         iy0 += ia5_0 * rx5;
         ry0 -= ia6_0 * ix6;
         iy0 += ia6_0 * rx6;
         ry0 -= ia7_0 * ix7;
         iy0 += ia7_0 * rx7;
      }
      for (i=M8; i < M; i += 2)
      {
         const register float rx0 = X[i], ix0 = X[i+1];
         const register float ra0_0=A0[i], ia0_0=A0[i+1];
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
      }
      #ifdef BETA0
         Y[0] = ry0;
         Y[1] = iy0;
      #else
         Y[0] += ry0;
         Y[1] += iy0;
      #endif
   }
}
