#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 12
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGEMV(ATL_CINT M0, ATL_CINT N, const TYPE *A, ATL_CINT lda0,
               const TYPE *X, TYPE *Y)
/*
 * 1x12 unrolled mvt_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N12=(N/12)*12, M=M0+M0, M1=((M0/1)*1)<<1,
            lda=lda0+lda0, lda12=lda*12;
   ATL_INT j;

   for (j=N12; j; j -= 12, A += lda12, Y += 24)
   {
      const double *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda,
                   *A5=A4+lda, *A6=A5+lda, *A7=A6+lda, *A8=A7+lda, *A9=A8+lda,
                   *A10=A9+lda, *A11=A10+lda;
      register double ry0=ATL_rzero, iy0=ATL_rzero, ry1=ATL_rzero,
                      iy1=ATL_rzero, ry2=ATL_rzero, iy2=ATL_rzero,
                      ry3=ATL_rzero, iy3=ATL_rzero, ry4=ATL_rzero,
                      iy4=ATL_rzero, ry5=ATL_rzero, iy5=ATL_rzero,
                      ry6=ATL_rzero, iy6=ATL_rzero, ry7=ATL_rzero,
                      iy7=ATL_rzero, ry8=ATL_rzero, iy8=ATL_rzero,
                      ry9=ATL_rzero, iy9=ATL_rzero, ry10=ATL_rzero,
                      iy10=ATL_rzero, ry11=ATL_rzero, iy11=ATL_rzero;
      ATL_INT i;
      prefY(Y+24+24-1);
      for (i=0; i < M1; i += 2)
      {
         const register double ra0_0=A0[i+0], ia0_0=A0[i+1], ra0_1=A1[i+0],
                               ia0_1=A1[i+1], ra0_2=A2[i+0], ia0_2=A2[i+1],
                               ra0_3=A3[i+0], ia0_3=A3[i+1], ra0_4=A4[i+0],
                               ia0_4=A4[i+1], ra0_5=A5[i+0], ia0_5=A5[i+1],
                               ra0_6=A6[i+0], ia0_6=A6[i+1], ra0_7=A7[i+0],
                               ia0_7=A7[i+1], ra0_8=A8[i+0], ia0_8=A8[i+1],
                               ra0_9=A9[i+0], ia0_9=A9[i+1], ra0_10=A10[i+0],
                               ia0_10=A10[i+1], ra0_11=A11[i+0],
                               ia0_11=A11[i+1];
         const register double rx0=X[i+0], ix0=X[i+1];

         prefX(X);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;
         ry1 += ra0_1 * rx0;
         iy1 += ra0_1 * ix0;
         ry2 += ra0_2 * rx0;
         iy2 += ra0_2 * ix0;
         ry3 += ra0_3 * rx0;
         iy3 += ra0_3 * ix0;
         ry4 += ra0_4 * rx0;
         iy4 += ra0_4 * ix0;
         ry5 += ra0_5 * rx0;
         iy5 += ra0_5 * ix0;
         ry6 += ra0_6 * rx0;
         iy6 += ra0_6 * ix0;
         ry7 += ra0_7 * rx0;
         iy7 += ra0_7 * ix0;
         ry8 += ra0_8 * rx0;
         iy8 += ra0_8 * ix0;
         ry9 += ra0_9 * rx0;
         iy9 += ra0_9 * ix0;
         ry10 += ra0_10 * rx0;
         iy10 += ra0_10 * ix0;
         ry11 += ra0_11 * rx0;
         iy11 += ra0_11 * ix0;

         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
         ry1 -= ia0_1 * ix0;
         iy1 += ia0_1 * rx0;
         prefA(A1);
         ry2 -= ia0_2 * ix0;
         iy2 += ia0_2 * rx0;
         prefA(A2);
         ry3 -= ia0_3 * ix0;
         iy3 += ia0_3 * rx0;
         prefA(A3);
         ry4 -= ia0_4 * ix0;
         iy4 += ia0_4 * rx0;
         prefA(A4);
         ry5 -= ia0_5 * ix0;
         iy5 += ia0_5 * rx0;
         prefA(A5);
         ry6 -= ia0_6 * ix0;
         iy6 += ia0_6 * rx0;
         prefA(A6);
         ry7 -= ia0_7 * ix0;
         iy7 += ia0_7 * rx0;
         prefA(A7);
         ry8 -= ia0_8 * ix0;
         iy8 += ia0_8 * rx0;
         prefA(A8);
         ry9 -= ia0_9 * ix0;
         iy9 += ia0_9 * rx0;
         prefA(A9);
         ry10 -= ia0_10 * ix0;
         iy10 += ia0_10 * rx0;
         prefA(A10);
         ry11 -= ia0_11 * ix0;
         iy11 += ia0_11 * rx0;
         prefA(A11);
      }
      #ifdef BETA0
         Y[0] = ry0;
         Y[1] = iy0;
         Y[2] = ry1;
         Y[3] = iy1;
         Y[4] = ry2;
         Y[5] = iy2;
         Y[6] = ry3;
         Y[7] = iy3;
         Y[8] = ry4;
         Y[9] = iy4;
         Y[10] = ry5;
         Y[11] = iy5;
         Y[12] = ry6;
         Y[13] = iy6;
         Y[14] = ry7;
         Y[15] = iy7;
         Y[16] = ry8;
         Y[17] = iy8;
         Y[18] = ry9;
         Y[19] = iy9;
         Y[20] = ry10;
         Y[21] = iy10;
         Y[22] = ry11;
         Y[23] = iy11;
      #else
         Y[0] += ry0;
         Y[1] += iy0;
         Y[2] += ry1;
         Y[3] += iy1;
         Y[4] += ry2;
         Y[5] += iy2;
         Y[6] += ry3;
         Y[7] += iy3;
         Y[8] += ry4;
         Y[9] += iy4;
         Y[10] += ry5;
         Y[11] += iy5;
         Y[12] += ry6;
         Y[13] += iy6;
         Y[14] += ry7;
         Y[15] += iy7;
         Y[16] += ry8;
         Y[17] += iy8;
         Y[18] += ry9;
         Y[19] += iy9;
         Y[20] += ry10;
         Y[21] += iy10;
         Y[22] += ry11;
         Y[23] += iy11;
      #endif
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N12; j; j--, A += lda, Y += 2)
   {
      const double *A0=A;
      register double ry0=ATL_rzero, iy0=ATL_rzero;
      ATL_INT i;
      prefY(Y+24+24-1);
      for (i=0; i < M1; i += 2)
      {
         const register double ra0_0=A0[i+0], ia0_0=A0[i+1];
         const register double rx0=X[i+0], ix0=X[i+1];

         prefX(X);
         ry0 += ra0_0 * rx0;
         iy0 += ra0_0 * ix0;

         ry0 -= ia0_0 * ix0;
         iy0 += ia0_0 * rx0;
         prefA(A0);
      }
      #ifdef BETA0
         Y[0] = ry0;
         Y[1] = iy0;
      #else
         Y[0] += ry0;
         Y[1] += iy0;
      #endif
   }
}
