#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 2
#endif
#ifndef PFXDIST
   #define PFXDIST 256
#endif
#ifndef PFADIST
   #define PFADIST 256
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGER2K(ATL_CINT M0, ATL_CINT N, const float *X, const float *Y, 
                const float *W, const float *Z, float *A, ATL_CINT lda0)
/*
 * 16x2 unrolled r2_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N2=(N/2)*2, M=M0+M0, M16=((M0/16)*16)<<1,
            lda=lda0+lda0, lda2=lda*2;
   ATL_INT j;

   for (j=N2; j; j -= 2, A += lda2, Y += 4)
   {
      float *A0=A, *A1=A0+lda;
      const register float ry0=Y[0], rz0=Z[0], iy0=Y[1], iz0=Z[1], ry1=Y[2],
                           rz1=Z[2], iy1=Y[3], iz1=Z[3];
      ATL_INT i;
      prefY(Y+4+4-1);
      for (i=0; i < M16; i += 32)
      {
         register float ra00, ia00, ra10, ia10;
         const float rx0=X[i+0], rw0=W[i+0], ix0=X[i+1], iw0=W[i+1], rx1=X[i+2],
                     rw1=W[i+2], ix1=X[i+3], iw1=W[i+3], rx2=X[i+4], rw2=W[i+4],
                     ix2=X[i+5], iw2=W[i+5], rx3=X[i+6], rw3=W[i+6], ix3=X[i+7],
                     iw3=W[i+7], rx4=X[i+8], rw4=W[i+8], ix4=X[i+9], iw4=W[i+9],
                     rx5=X[i+10], rw5=W[i+10], ix5=X[i+11], iw5=W[i+11],
                     rx6=X[i+12], rw6=W[i+12], ix6=X[i+13], iw6=W[i+13],
                     rx7=X[i+14], rw7=W[i+14], ix7=X[i+15], iw7=W[i+15],
                     rx8=X[i+16], rw8=W[i+16], ix8=X[i+17], iw8=W[i+17],
                     rx9=X[i+18], rw9=W[i+18], ix9=X[i+19], iw9=W[i+19],
                     rx10=X[i+20], rw10=W[i+20], ix10=X[i+21], iw10=W[i+21],
                     rx11=X[i+22], rw11=W[i+22], ix11=X[i+23], iw11=W[i+23],
                     rx12=X[i+24], rw12=W[i+24], ix12=X[i+25], iw12=W[i+25],
                     rx13=X[i+26], rw13=W[i+26], ix13=X[i+27], iw13=W[i+27],
                     rx14=X[i+28], rw14=W[i+28], ix14=X[i+29], iw14=W[i+29],
                     rx15=X[i+30], rw15=W[i+30], ix15=X[i+31], iw15=W[i+31];

         prefX(X);
         ra00 = A0[i+0];
         ia00 = A0[i+0+1];
         ra10 = A0[i+0+2];
         ia10 = A0[i+0+3];
         ra00 += rx0 * ry0;
         ia00 += rx0 * iy0;
         ra10 += rx1 * ry0;
         ia10 += rx1 * iy0;
         prefA(A0);
         ra00 -= ix0 * iy0;
         ia00 += ix0 * ry0;
         ra10 -= ix1 * iy0;
         ia10 += ix1 * ry0;
         prefX(W);
         ra00 += rw0 * rz0;
         ia00 += rw0 * iz0;
         ra10 += rw1 * rz0;
         ia10 += rw1 * iz0;
         ra00 -= iw0 * iz0;
         A0[i+0] = ra00;
         ia00 += iw0 * rz0;
         A0[i+0+1] = ia00;
         ra10 -= iw1 * iz0;
         A0[i+0+2] = ra10;
         ia10 += iw1 * rz0;
         A0[i+0+3] = ia10;
         ra00 = A0[i+4];
         ia00 = A0[i+4+1];
         ra10 = A0[i+4+2];
         ia10 = A0[i+4+3];
         ra00 += rx2 * ry0;
         ia00 += rx2 * iy0;
         ra10 += rx3 * ry0;
         ia10 += rx3 * iy0;
         ra00 -= ix2 * iy0;
         ia00 += ix2 * ry0;
         ra10 -= ix3 * iy0;
         ia10 += ix3 * ry0;
         ra00 += rw2 * rz0;
         ia00 += rw2 * iz0;
         ra10 += rw3 * rz0;
         ia10 += rw3 * iz0;
         ra00 -= iw2 * iz0;
         A0[i+4] = ra00;
         ia00 += iw2 * rz0;
         A0[i+4+1] = ia00;
         ra10 -= iw3 * iz0;
         A0[i+4+2] = ra10;
         ia10 += iw3 * rz0;
         A0[i+4+3] = ia10;
         ra00 = A0[i+8];
         ia00 = A0[i+8+1];
         ra10 = A0[i+8+2];
         ia10 = A0[i+8+3];
         ra00 += rx4 * ry0;
         ia00 += rx4 * iy0;
         ra10 += rx5 * ry0;
         ia10 += rx5 * iy0;
         ra00 -= ix4 * iy0;
         ia00 += ix4 * ry0;
         ra10 -= ix5 * iy0;
         ia10 += ix5 * ry0;
         ra00 += rw4 * rz0;
         ia00 += rw4 * iz0;
         ra10 += rw5 * rz0;
         ia10 += rw5 * iz0;
         ra00 -= iw4 * iz0;
         A0[i+8] = ra00;
         ia00 += iw4 * rz0;
         A0[i+8+1] = ia00;
         ra10 -= iw5 * iz0;
         A0[i+8+2] = ra10;
         ia10 += iw5 * rz0;
         A0[i+8+3] = ia10;
         ra00 = A0[i+12];
         ia00 = A0[i+12+1];
         ra10 = A0[i+12+2];
         ia10 = A0[i+12+3];
         ra00 += rx6 * ry0;
         ia00 += rx6 * iy0;
         ra10 += rx7 * ry0;
         ia10 += rx7 * iy0;
         ra00 -= ix6 * iy0;
         ia00 += ix6 * ry0;
         ra10 -= ix7 * iy0;
         ia10 += ix7 * ry0;
         ra00 += rw6 * rz0;
         ia00 += rw6 * iz0;
         ra10 += rw7 * rz0;
         ia10 += rw7 * iz0;
         ra00 -= iw6 * iz0;
         A0[i+12] = ra00;
         ia00 += iw6 * rz0;
         A0[i+12+1] = ia00;
         ra10 -= iw7 * iz0;
         A0[i+12+2] = ra10;
         ia10 += iw7 * rz0;
         A0[i+12+3] = ia10;
         ra00 = A0[i+16];
         ia00 = A0[i+16+1];
         ra10 = A0[i+16+2];
         ia10 = A0[i+16+3];
         ra00 += rx8 * ry0;
         ia00 += rx8 * iy0;
         ra10 += rx9 * ry0;
         ia10 += rx9 * iy0;
         prefA(A0);
         ra00 -= ix8 * iy0;
         ia00 += ix8 * ry0;
         ra10 -= ix9 * iy0;
         ia10 += ix9 * ry0;
         ra00 += rw8 * rz0;
         ia00 += rw8 * iz0;
         ra10 += rw9 * rz0;
         ia10 += rw9 * iz0;
         ra00 -= iw8 * iz0;
         A0[i+16] = ra00;
         ia00 += iw8 * rz0;
         A0[i+16+1] = ia00;
         ra10 -= iw9 * iz0;
         A0[i+16+2] = ra10;
         ia10 += iw9 * rz0;
         A0[i+16+3] = ia10;
         ra00 = A0[i+20];
         ia00 = A0[i+20+1];
         ra10 = A0[i+20+2];
         ia10 = A0[i+20+3];
         ra00 += rx10 * ry0;
         ia00 += rx10 * iy0;
         ra10 += rx11 * ry0;
         ia10 += rx11 * iy0;
         ra00 -= ix10 * iy0;
         ia00 += ix10 * ry0;
         ra10 -= ix11 * iy0;
         ia10 += ix11 * ry0;
         ra00 += rw10 * rz0;
         ia00 += rw10 * iz0;
         ra10 += rw11 * rz0;
         ia10 += rw11 * iz0;
         ra00 -= iw10 * iz0;
         A0[i+20] = ra00;
         ia00 += iw10 * rz0;
         A0[i+20+1] = ia00;
         ra10 -= iw11 * iz0;
         A0[i+20+2] = ra10;
         ia10 += iw11 * rz0;
         A0[i+20+3] = ia10;
         ra00 = A0[i+24];
         ia00 = A0[i+24+1];
         ra10 = A0[i+24+2];
         ia10 = A0[i+24+3];
         ra00 += rx12 * ry0;
         ia00 += rx12 * iy0;
         ra10 += rx13 * ry0;
         ia10 += rx13 * iy0;
         ra00 -= ix12 * iy0;
         ia00 += ix12 * ry0;
         ra10 -= ix13 * iy0;
         ia10 += ix13 * ry0;
         ra00 += rw12 * rz0;
         ia00 += rw12 * iz0;
         ra10 += rw13 * rz0;
         ia10 += rw13 * iz0;
         ra00 -= iw12 * iz0;
         A0[i+24] = ra00;
         ia00 += iw12 * rz0;
         A0[i+24+1] = ia00;
         ra10 -= iw13 * iz0;
         A0[i+24+2] = ra10;
         ia10 += iw13 * rz0;
         A0[i+24+3] = ia10;
         ra00 = A0[i+28];
         ia00 = A0[i+28+1];
         ra10 = A0[i+28+2];
         ia10 = A0[i+28+3];
         ra00 += rx14 * ry0;
         ia00 += rx14 * iy0;
         ra10 += rx15 * ry0;
         ia10 += rx15 * iy0;
         ra00 -= ix14 * iy0;
         ia00 += ix14 * ry0;
         ra10 -= ix15 * iy0;
         ia10 += ix15 * ry0;
         ra00 += rw14 * rz0;
         ia00 += rw14 * iz0;
         ra10 += rw15 * rz0;
         ia10 += rw15 * iz0;
         ra00 -= iw14 * iz0;
         A0[i+28] = ra00;
         ia00 += iw14 * rz0;
         A0[i+28+1] = ia00;
         ra10 -= iw15 * iz0;
         A0[i+28+2] = ra10;
         ia10 += iw15 * rz0;
         A0[i+28+3] = ia10;

         ra00 = A1[i+0];
         ia00 = A1[i+0+1];
         ra10 = A1[i+0+2];
         ia10 = A1[i+0+3];
         ra00 += rx0 * ry1;
         ia00 += rx0 * iy1;
         ra10 += rx1 * ry1;
         ia10 += rx1 * iy1;
         prefA(A1);
         ra00 -= ix0 * iy1;
         ia00 += ix0 * ry1;
         ra10 -= ix1 * iy1;
         ia10 += ix1 * ry1;
         ra00 += rw0 * rz1;
         ia00 += rw0 * iz1;
         ra10 += rw1 * rz1;
         ia10 += rw1 * iz1;
         ra00 -= iw0 * iz1;
         A1[i+0] = ra00;
         ia00 += iw0 * rz1;
         A1[i+0+1] = ia00;
         ra10 -= iw1 * iz1;
         A1[i+0+2] = ra10;
         ia10 += iw1 * rz1;
         A1[i+0+3] = ia10;
         ra00 = A1[i+4];
         ia00 = A1[i+4+1];
         ra10 = A1[i+4+2];
         ia10 = A1[i+4+3];
         ra00 += rx2 * ry1;
         ia00 += rx2 * iy1;
         ra10 += rx3 * ry1;
         ia10 += rx3 * iy1;
         ra00 -= ix2 * iy1;
         ia00 += ix2 * ry1;
         ra10 -= ix3 * iy1;
         ia10 += ix3 * ry1;
         ra00 += rw2 * rz1;
         ia00 += rw2 * iz1;
         ra10 += rw3 * rz1;
         ia10 += rw3 * iz1;
         ra00 -= iw2 * iz1;
         A1[i+4] = ra00;
         ia00 += iw2 * rz1;
         A1[i+4+1] = ia00;
         ra10 -= iw3 * iz1;
         A1[i+4+2] = ra10;
         ia10 += iw3 * rz1;
         A1[i+4+3] = ia10;
         ra00 = A1[i+8];
         ia00 = A1[i+8+1];
         ra10 = A1[i+8+2];
         ia10 = A1[i+8+3];
         ra00 += rx4 * ry1;
         ia00 += rx4 * iy1;
         ra10 += rx5 * ry1;
         ia10 += rx5 * iy1;
         ra00 -= ix4 * iy1;
         ia00 += ix4 * ry1;
         ra10 -= ix5 * iy1;
         ia10 += ix5 * ry1;
         ra00 += rw4 * rz1;
         ia00 += rw4 * iz1;
         ra10 += rw5 * rz1;
         ia10 += rw5 * iz1;
         ra00 -= iw4 * iz1;
         A1[i+8] = ra00;
         ia00 += iw4 * rz1;
         A1[i+8+1] = ia00;
         ra10 -= iw5 * iz1;
         A1[i+8+2] = ra10;
         ia10 += iw5 * rz1;
         A1[i+8+3] = ia10;
         ra00 = A1[i+12];
         ia00 = A1[i+12+1];
         ra10 = A1[i+12+2];
         ia10 = A1[i+12+3];
         ra00 += rx6 * ry1;
         ia00 += rx6 * iy1;
         ra10 += rx7 * ry1;
         ia10 += rx7 * iy1;
         ra00 -= ix6 * iy1;
         ia00 += ix6 * ry1;
         ra10 -= ix7 * iy1;
         ia10 += ix7 * ry1;
         ra00 += rw6 * rz1;
         ia00 += rw6 * iz1;
         ra10 += rw7 * rz1;
         ia10 += rw7 * iz1;
         ra00 -= iw6 * iz1;
         A1[i+12] = ra00;
         ia00 += iw6 * rz1;
         A1[i+12+1] = ia00;
         ra10 -= iw7 * iz1;
         A1[i+12+2] = ra10;
         ia10 += iw7 * rz1;
         A1[i+12+3] = ia10;
         ra00 = A1[i+16];
         ia00 = A1[i+16+1];
         ra10 = A1[i+16+2];
         ia10 = A1[i+16+3];
         ra00 += rx8 * ry1;
         ia00 += rx8 * iy1;
         ra10 += rx9 * ry1;
         ia10 += rx9 * iy1;
         prefA(A1);
         ra00 -= ix8 * iy1;
         ia00 += ix8 * ry1;
         ra10 -= ix9 * iy1;
         ia10 += ix9 * ry1;
         ra00 += rw8 * rz1;
         ia00 += rw8 * iz1;
         ra10 += rw9 * rz1;
         ia10 += rw9 * iz1;
         ra00 -= iw8 * iz1;
         A1[i+16] = ra00;
         ia00 += iw8 * rz1;
         A1[i+16+1] = ia00;
         ra10 -= iw9 * iz1;
         A1[i+16+2] = ra10;
         ia10 += iw9 * rz1;
         A1[i+16+3] = ia10;
         ra00 = A1[i+20];
         ia00 = A1[i+20+1];
         ra10 = A1[i+20+2];
         ia10 = A1[i+20+3];
         ra00 += rx10 * ry1;
         ia00 += rx10 * iy1;
         ra10 += rx11 * ry1;
         ia10 += rx11 * iy1;
         ra00 -= ix10 * iy1;
         ia00 += ix10 * ry1;
         ra10 -= ix11 * iy1;
         ia10 += ix11 * ry1;
         ra00 += rw10 * rz1;
         ia00 += rw10 * iz1;
         ra10 += rw11 * rz1;
         ia10 += rw11 * iz1;
         ra00 -= iw10 * iz1;
         A1[i+20] = ra00;
         ia00 += iw10 * rz1;
         A1[i+20+1] = ia00;
         ra10 -= iw11 * iz1;
         A1[i+20+2] = ra10;
         ia10 += iw11 * rz1;
         A1[i+20+3] = ia10;
         ra00 = A1[i+24];
         ia00 = A1[i+24+1];
         ra10 = A1[i+24+2];
         ia10 = A1[i+24+3];
         ra00 += rx12 * ry1;
         ia00 += rx12 * iy1;
         ra10 += rx13 * ry1;
         ia10 += rx13 * iy1;
         ra00 -= ix12 * iy1;
         ia00 += ix12 * ry1;
         ra10 -= ix13 * iy1;
         ia10 += ix13 * ry1;
         ra00 += rw12 * rz1;
         ia00 += rw12 * iz1;
         ra10 += rw13 * rz1;
         ia10 += rw13 * iz1;
         ra00 -= iw12 * iz1;
         A1[i+24] = ra00;
         ia00 += iw12 * rz1;
         A1[i+24+1] = ia00;
         ra10 -= iw13 * iz1;
         A1[i+24+2] = ra10;
         ia10 += iw13 * rz1;
         A1[i+24+3] = ia10;
         ra00 = A1[i+28];
         ia00 = A1[i+28+1];
         ra10 = A1[i+28+2];
         ia10 = A1[i+28+3];
         ra00 += rx14 * ry1;
         ia00 += rx14 * iy1;
         ra10 += rx15 * ry1;
         ia10 += rx15 * iy1;
         ra00 -= ix14 * iy1;
         ia00 += ix14 * ry1;
         ra10 -= ix15 * iy1;
         ia10 += ix15 * ry1;
         ra00 += rw14 * rz1;
         ia00 += rw14 * iz1;
         ra10 += rw15 * rz1;
         ia10 += rw15 * iz1;
         ra00 -= iw14 * iz1;
         A1[i+28] = ra00;
         ia00 += iw14 * rz1;
         A1[i+28+1] = ia00;
         ra10 -= iw15 * iz1;
         A1[i+28+2] = ra10;
         ia10 += iw15 * rz1;
         A1[i+28+3] = ia10;

      }
      for (i=M16; i < M; i += 2)
      {
         const float rx0 = X[i], ix0 = X[i+1], rw0 = W[i], iw0 = W[i+1];
         A0[i] += rx0 * ry0 - ix0 * iy0 + rw0 * rz0 - iw0 * iz0 ;
         A0[i+1] += ix0 * ry0 + rx0 * iy0 + iw0 * rz0 + rw0 * iz0;
         A1[i] += rx0 * ry1 - ix0 * iy1 + rw0 * rz1 - iw0 * iz1 ;
         A1[i+1] += ix0 * ry1 + rx0 * iy1 + iw0 * rz1 + rw0 * iz1;
      }
      Z += 4;
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N2; j; j--, A += lda, Y += 2)
   {
      float *A0=A;
      const register float ry0=Y[0], rz0=Z[0], iy0=Y[1], iz0=Z[1];
      ATL_INT i;
      prefY(Y+4+4-1);
      for (i=0; i < M16; i += 32)
      {
         register float ra00, ia00, ra10, ia10;
         const float rx0=X[i+0], rw0=W[i+0], ix0=X[i+1], iw0=W[i+1], rx1=X[i+2],
                     rw1=W[i+2], ix1=X[i+3], iw1=W[i+3], rx2=X[i+4], rw2=W[i+4],
                     ix2=X[i+5], iw2=W[i+5], rx3=X[i+6], rw3=W[i+6], ix3=X[i+7],
                     iw3=W[i+7], rx4=X[i+8], rw4=W[i+8], ix4=X[i+9], iw4=W[i+9],
                     rx5=X[i+10], rw5=W[i+10], ix5=X[i+11], iw5=W[i+11],
                     rx6=X[i+12], rw6=W[i+12], ix6=X[i+13], iw6=W[i+13],
                     rx7=X[i+14], rw7=W[i+14], ix7=X[i+15], iw7=W[i+15],
                     rx8=X[i+16], rw8=W[i+16], ix8=X[i+17], iw8=W[i+17],
                     rx9=X[i+18], rw9=W[i+18], ix9=X[i+19], iw9=W[i+19],
                     rx10=X[i+20], rw10=W[i+20], ix10=X[i+21], iw10=W[i+21],
                     rx11=X[i+22], rw11=W[i+22], ix11=X[i+23], iw11=W[i+23],
                     rx12=X[i+24], rw12=W[i+24], ix12=X[i+25], iw12=W[i+25],
                     rx13=X[i+26], rw13=W[i+26], ix13=X[i+27], iw13=W[i+27],
                     rx14=X[i+28], rw14=W[i+28], ix14=X[i+29], iw14=W[i+29],
                     rx15=X[i+30], rw15=W[i+30], ix15=X[i+31], iw15=W[i+31];

         prefX(X);
         ra00 = A0[i+0];
         ia00 = A0[i+0+1];
         ra10 = A0[i+0+2];
         ia10 = A0[i+0+3];
         ra00 += rx0 * ry0;
         ia00 += rx0 * iy0;
         ra10 += rx1 * ry0;
         ia10 += rx1 * iy0;
         prefA(A0);
         ra00 -= ix0 * iy0;
         ia00 += ix0 * ry0;
         ra10 -= ix1 * iy0;
         ia10 += ix1 * ry0;
         prefX(W);
         ra00 += rw0 * rz0;
         ia00 += rw0 * iz0;
         ra10 += rw1 * rz0;
         ia10 += rw1 * iz0;
         ra00 -= iw0 * iz0;
         A0[i+0] = ra00;
         ia00 += iw0 * rz0;
         A0[i+0+1] = ia00;
         ra10 -= iw1 * iz0;
         A0[i+0+2] = ra10;
         ia10 += iw1 * rz0;
         A0[i+0+3] = ia10;
         ra00 = A0[i+4];
         ia00 = A0[i+4+1];
         ra10 = A0[i+4+2];
         ia10 = A0[i+4+3];
         ra00 += rx2 * ry0;
         ia00 += rx2 * iy0;
         ra10 += rx3 * ry0;
         ia10 += rx3 * iy0;
         ra00 -= ix2 * iy0;
         ia00 += ix2 * ry0;
         ra10 -= ix3 * iy0;
         ia10 += ix3 * ry0;
         ra00 += rw2 * rz0;
         ia00 += rw2 * iz0;
         ra10 += rw3 * rz0;
         ia10 += rw3 * iz0;
         ra00 -= iw2 * iz0;
         A0[i+4] = ra00;
         ia00 += iw2 * rz0;
         A0[i+4+1] = ia00;
         ra10 -= iw3 * iz0;
         A0[i+4+2] = ra10;
         ia10 += iw3 * rz0;
         A0[i+4+3] = ia10;
         ra00 = A0[i+8];
         ia00 = A0[i+8+1];
         ra10 = A0[i+8+2];
         ia10 = A0[i+8+3];
         ra00 += rx4 * ry0;
         ia00 += rx4 * iy0;
         ra10 += rx5 * ry0;
         ia10 += rx5 * iy0;
         ra00 -= ix4 * iy0;
         ia00 += ix4 * ry0;
         ra10 -= ix5 * iy0;
         ia10 += ix5 * ry0;
         ra00 += rw4 * rz0;
         ia00 += rw4 * iz0;
         ra10 += rw5 * rz0;
         ia10 += rw5 * iz0;
         ra00 -= iw4 * iz0;
         A0[i+8] = ra00;
         ia00 += iw4 * rz0;
         A0[i+8+1] = ia00;
         ra10 -= iw5 * iz0;
         A0[i+8+2] = ra10;
         ia10 += iw5 * rz0;
         A0[i+8+3] = ia10;
         ra00 = A0[i+12];
         ia00 = A0[i+12+1];
         ra10 = A0[i+12+2];
         ia10 = A0[i+12+3];
         ra00 += rx6 * ry0;
         ia00 += rx6 * iy0;
         ra10 += rx7 * ry0;
         ia10 += rx7 * iy0;
         ra00 -= ix6 * iy0;
         ia00 += ix6 * ry0;
         ra10 -= ix7 * iy0;
         ia10 += ix7 * ry0;
         ra00 += rw6 * rz0;
         ia00 += rw6 * iz0;
         ra10 += rw7 * rz0;
         ia10 += rw7 * iz0;
         ra00 -= iw6 * iz0;
         A0[i+12] = ra00;
         ia00 += iw6 * rz0;
         A0[i+12+1] = ia00;
         ra10 -= iw7 * iz0;
         A0[i+12+2] = ra10;
         ia10 += iw7 * rz0;
         A0[i+12+3] = ia10;
         ra00 = A0[i+16];
         ia00 = A0[i+16+1];
         ra10 = A0[i+16+2];
         ia10 = A0[i+16+3];
         ra00 += rx8 * ry0;
         ia00 += rx8 * iy0;
         ra10 += rx9 * ry0;
         ia10 += rx9 * iy0;
         prefA(A0);
         ra00 -= ix8 * iy0;
         ia00 += ix8 * ry0;
         ra10 -= ix9 * iy0;
         ia10 += ix9 * ry0;
         ra00 += rw8 * rz0;
         ia00 += rw8 * iz0;
         ra10 += rw9 * rz0;
         ia10 += rw9 * iz0;
         ra00 -= iw8 * iz0;
         A0[i+16] = ra00;
         ia00 += iw8 * rz0;
         A0[i+16+1] = ia00;
         ra10 -= iw9 * iz0;
         A0[i+16+2] = ra10;
         ia10 += iw9 * rz0;
         A0[i+16+3] = ia10;
         ra00 = A0[i+20];
         ia00 = A0[i+20+1];
         ra10 = A0[i+20+2];
         ia10 = A0[i+20+3];
         ra00 += rx10 * ry0;
         ia00 += rx10 * iy0;
         ra10 += rx11 * ry0;
         ia10 += rx11 * iy0;
         ra00 -= ix10 * iy0;
         ia00 += ix10 * ry0;
         ra10 -= ix11 * iy0;
         ia10 += ix11 * ry0;
         ra00 += rw10 * rz0;
         ia00 += rw10 * iz0;
         ra10 += rw11 * rz0;
         ia10 += rw11 * iz0;
         ra00 -= iw10 * iz0;
         A0[i+20] = ra00;
         ia00 += iw10 * rz0;
         A0[i+20+1] = ia00;
         ra10 -= iw11 * iz0;
         A0[i+20+2] = ra10;
         ia10 += iw11 * rz0;
         A0[i+20+3] = ia10;
         ra00 = A0[i+24];
         ia00 = A0[i+24+1];
         ra10 = A0[i+24+2];
         ia10 = A0[i+24+3];
         ra00 += rx12 * ry0;
         ia00 += rx12 * iy0;
         ra10 += rx13 * ry0;
         ia10 += rx13 * iy0;
         ra00 -= ix12 * iy0;
         ia00 += ix12 * ry0;
         ra10 -= ix13 * iy0;
         ia10 += ix13 * ry0;
         ra00 += rw12 * rz0;
         ia00 += rw12 * iz0;
         ra10 += rw13 * rz0;
         ia10 += rw13 * iz0;
         ra00 -= iw12 * iz0;
         A0[i+24] = ra00;
         ia00 += iw12 * rz0;
         A0[i+24+1] = ia00;
         ra10 -= iw13 * iz0;
         A0[i+24+2] = ra10;
         ia10 += iw13 * rz0;
         A0[i+24+3] = ia10;
         ra00 = A0[i+28];
         ia00 = A0[i+28+1];
         ra10 = A0[i+28+2];
         ia10 = A0[i+28+3];
         ra00 += rx14 * ry0;
         ia00 += rx14 * iy0;
         ra10 += rx15 * ry0;
         ia10 += rx15 * iy0;
         ra00 -= ix14 * iy0;
         ia00 += ix14 * ry0;
         ra10 -= ix15 * iy0;
         ia10 += ix15 * ry0;
         ra00 += rw14 * rz0;
         ia00 += rw14 * iz0;
         ra10 += rw15 * rz0;
         ia10 += rw15 * iz0;
         ra00 -= iw14 * iz0;
         A0[i+28] = ra00;
         ia00 += iw14 * rz0;
         A0[i+28+1] = ia00;
         ra10 -= iw15 * iz0;
         A0[i+28+2] = ra10;
         ia10 += iw15 * rz0;
         A0[i+28+3] = ia10;

      }
      for (i=M16; i < M; i += 2)
      {
         const float rx0 = X[i], ix0 = X[i+1], rw0 = W[i], iw0 = W[i+1];
         A0[i] += rx0 * ry0 - ix0 * iy0 + rw0 * rz0 - iw0 * iz0 ;
         A0[i+1] += ix0 * ry0 + rx0 * iy0 + iw0 * rz0 + rw0 * iz0;
      }
      Z += 2;
   }
}
